//
// (c)2019 Sassafras Software Inc.
//

// #mark TD
// #mark -

function TD (url, appid, opts) {
	opts = opts || {};
	if (!url)
		url = '';
	if (!url.endsWith('/'))
		url += '/';
	if (opts.sandbox)
		url += "SB";
	this.url = url + "TDWebApi/api/";
	this.appid = appid || 30;
	this.lookups = {};
	this.config = bundle.load("td-lookups.json");
	if (opts.creds)
		this.auth(opts.creds);
	this.verbose = !!opts.verbose;
	return(this);
}

(function (proto) {

	proto.auth = function (creds) {
		if (!creds) {
			if (this.token)
				return({ token:this.token, expires:this.expires });
			return(null);
		}

		if (creds.token) {
			this.token = creds.token;
			this.expires = new Date(creds.expires);
			if (this.expires.getTime() - 7200000 < Date.now()) {
				this.token = null;
				this.expires = null;
			} else {
				var result = this.ajax({ endpoint: "applications" });
				if (!result.value) {
					this.token = null;
					this.expires = null;
				}
			}
			return(this);
		}

		if (this.verbose) {
			print("--------------------------------");
			print("REQUEST: AUTH");
			print("POST " + this.url + "auth" + (creds.WebServicesKey ? "/loginadmin" : ""));
			print(JSON.stringify(creds));
			print("");
		}

		var xhr = new XMLHttpRequest();
		xhr.open("POST", this.url + "auth" + (creds.WebServicesKey ? "/loginadmin" : ""), false);
		xhr.timeout = 15000;
		xhr.setRequestHeader("Content-Type", "application/json");
		xhr.send(JSON.stringify(creds));
		if (xhr.status == 200) {
			var parts = xhr.responseText.split('.');
			if (parts.length == 3) {
				this.jwt = {
						header:JSON.parse(decode(parts[0])),
						payload:JSON.parse(decode(parts[1])),
						sig:parts[2]
					};
				this.token = xhr.responseText;
				this.expires = new Date(this.jwt.payload.exp * 1000);
			}
		}

		if (this.verbose) {
			print("RESPONSE: AUTH");
			print(xhr.status + " " + xhr.statusText);
			print(xhr.responseText);
			print("");
		}

		return(this);
	}

	proto.verbosity = function (verbosity) {
		this.verbose = verbosity;
		return(this);
	}

	proto.log = function (desc) {
		if (desc.failure)
			console.error(desc.failure);
		return(desc);
	}

	proto.citems = function (keys, screen) {
		return(new TDCItems(this, keys, screen));
	}

	proto.assets = function (keys, screen) {
		return(new TDAssets(this, keys, screen));
	}

	proto.lookup = function (endpoint, key) {
		var known = this.config[endpoint];
		if (known) {
			if (typeof known === "string") {
				endpoint = known;
			} else {
				endpoint = known.endpoint;
				key = known.key;
			}
		}
		if (!this.lookups[endpoint])
			this.lookups[endpoint] = new TDLookup(this, endpoint.replace('#', this.appid), key);
		return(this.lookups[endpoint]);
	}

	proto.known = function () {
		var list = [];
		for (var key in this.config)
			list.push(key);
		return(list);
	}

	proto.ajax = function (opts) {
		var xhr, maxtries = 5, start;

		if (this.verbose) {
			start = new Date();
			print("--------------------------------");
			print("REQUEST: " + start.toISOString());
			print((opts.method || (opts.data ? "POST" : "GET")) + " " + opts.endpoint);
			print(opts.data ? JSON.stringify(opts.data) : "<empty>");
			print("");
		}

		while (!xhr) {
			xhr = new XMLHttpRequest();
			xhr.open(opts.method || (opts.data ? "POST" : "GET"), this.url + opts.endpoint, false);
			xhr.timeout = opts.timeout || 600000;
			if (this.token)
				xhr.setRequestHeader("Authorization", "Bearer " + this.token);
			xhr.setRequestHeader("Accept", "application/json");
			if (!opts.data) {
				xhr.send();
			} else {
				xhr.setRequestHeader("Content-Type", "application/json");
				xhr.send(JSON.stringify(opts.data));
			}
			if (--maxtries > 0 && xhr.status == 429) {
				var deadline = rfc1123date(xhr.getResponseHeader("X-RateLimit-Reset")),
					delay = 60;
				xhr = null;
				if (deadline) {
					delay = Math.floor((deadline.getTime() - Date.now()) / 1000) + 5;
					if (delay < 0)
						delay = 5;
				}
				if (this.verbose)
					print("rate limited until " + deadline + ", sleeping for " + delay + " seconds");
				console.info("rate limited until " + deadline + ", sleeping for " + delay + " seconds");
				sleep(delay * 1000);
				start = new Date();
			}
		}

		if (this.verbose) {
			var stop = new Date();
			var elapsed = (stop.getTime() - start.getTime()) / 1000;
			print("RESPONSE: " + stop.toISOString() + " -- elapsed: " + elapsed);
			print(xhr.status + " " + xhr.statusText);
			print(xhr.responseText);
			print("");
		}

		if (xhr.status != 200 && xhr.status != 201)
			return(this.log({
				failure:"TeamDynamix REST API call failed [" + opts.endpoint + "] (" + xhr.status + ": " + xhr.statusText + ")",
				status:xhr.status,
				data:xhr.response || xhr.responseText,
				opts:opts
			}));

		var result = xhr.response;
		if (!result)
			return(this.log({
				failure:"failed to parse TeamDynamix REST API response [" + opts.endpoint + "]",
				status:xhr.status,
				data:xhr.responseText,
				opts:opts
			}));

		return({ value:result, status:xhr.status });
	}

	function rfc1123date (s) {
		var months = [ "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" ],
			date = s.split(" "),
			time = date[4].split(":");
		date[2] = months.indexOf(date[2]);
		if (date[2] < 0)
			return(null);

		var when = new Date(parseInt(date[3]), date[2], parseInt(date[1]), parseInt(time[0]), parseInt(time[1]), parseInt(time[2]));
		return(new Date(when.getTime() - when.getTimezoneOffset() * 60000));
	}

})(TD.prototype);

// #mark -
// #mark TDTable
// #mark -

function TDTable (td, name, component, keys, screen) {
	this.td = td;
	this.list = [];
	this.fields = [];
	this.musts = {};
	this.idmap = {};
	this.name = name;
	this.component = component;
	this.screen = screen;
	this.keys = keys;

	return(this);
}

(function (proto) {

	proto.load = function (force) {
		if (!this.loaded || force) {
			var result = this.td.ajax({
				endpoint: this.td.appid + "/" + this.name + "/search",
				data: this.filter(typeof this.screen !== "function" && this.screen)
			});
			if (result.value) {
				this.idmap = {};
				if (typeof this.screen !== "function") {
					this.list = result.value;
					for (var i = 0; i < this.list.length; i++)
						this.map(this.list[i]);
				} else {
					this.list = [];
					for (var i = 0; i < result.value.length; i++) {
						var item = this.screen(result.value[i]);
						if (item) {
							this.list.push(item);
							this.map(item);
						}
					}
				}
				this.loaded = true;
			}
		}

		return(this);
	}

	proto.get = function (id) {
		if (id) {
			var result = this.td.ajax({
				endpoint: this.td.appid + "/" + this.name + "/" + id
			});
			return(result.value);
		}
	}

	proto.forEach = function (cb) {
		app.progress.open(this.list.length);
		for (var i = 0; i < this.list.length; i++) {
			if (cb(this.list[i]))
				break;
			app.progress.advance();
		}
		app.progress.close();
		return(this);
	}

	proto.commit = function (one) {
		return(this);
	}

	proto.filter = function (base) {
		var filter = base || {};
		filter.IsActive = true;
		return(filter);
	}

	proto.map = function (item) {
		if (item.ID)
			this.idmap[item.ID] = item;

		var keys = this.keys || [];
		for (var i = 0; i < keys.length; i++) {
			var value = item[keys[i]];
			if (value)
				this.idmap[value] = item;
		}
	}

	proto.find = function (id) {
		if (typeof id !== "object")
			return(id && this.idmap[id]);

		var item,
			keys = this.keys || [];
		for (var i = 0; i < keys.length; i++) {
			var value = id[keys[i]];
			if (value) {
				item = this.idmap[value];
				if (item)
					break;
			}
		}
		return(item);
	}

	proto.add = function (item, update) {
		if (item) {
			var match = this.find(item);
			if (match)
				item = extend(match, item);
			else
				this.list.push(item);
			this.map(item);
			this.commit(item, update);
		}
		return(item);
	}

	proto.setfields = function (fields) {
		this.fields = fields;
		return(this);
	}

	proto.must = function (field, value) {
		this.musts[field] = value;
		return(this);
	}

	proto.getattrs = function (pick) {
		var result = this.td.ajax({ endpoint:"attributes/custom?componentId=" + this.component + "&appId=" + this.td.appid }),
			attrs = {};
		if (result.value) {
			for (var i = 0; i < result.value.length; i++) {
				var attr = result.value[i];
				if (this._checkattr(attr)) {
					attrs[attr.Name] = attr;
					attrs[attr.ID] = attr;
				}
			}
		}
		return(pick ? attrs[pick] : attrs);
	}

	proto._tdify = function (imp, dst, src, otherattr) {
		var haveattr = {};
		dst = dst || extend({}, src, { Attachments:1, Uri:1, attributes:1, dirty:1 });
		for (var i = 0; i < imp.mappings.length; i++) {
			var field = imp.mappings[i],
				attrid = field.CustomAttributeID;
			if (!attrid) {
				dst[field.FieldIdentifier] = src[field.FieldIdentifier];
			} else if (src.attributes) {
				if (!dst.Attributes)
					dst.Attributes = [];
				dst.Attributes.push({
					ID:attrid,
					Value:src.attributes[imp.attrmap[attrid]]
				});
				haveattr[attrid] = true;
			}
		}
		for (var key in this.musts) {
			if (!dst[key])
				dst[key] = src[key] || this.musts[key];
		}
		if (otherattr) {
			for (var i = 0; i < otherattr.length; i++) {
				var oat = otherattr[i];
				if (haveattr[oat.ID])
					continue;
				if (src.attributes && oat.Name && oat.Name.indexOf("KS:") == 0)
					continue;
				if (!dst.Attributes)
					dst.Attributes = [];
				dst.Attributes.push({
					ID:oat.ID,
					Value:oat.Value
				});
			}
		}
		return(dst);
	}

	proto._importer = function (mappings, items) {
		return(this._impcheck(this.td.ajax({
				endpoint: this.td.appid + "/" + this.name + "/import",
				data: {
					Settings:{
						Mappings:mappings,
						CreateItems:true,
						UpdateItems:true
					},
					Items:items
				}
			})));
	}

	proto._impcheck = function (result) {
		var list = result.value || [];
		for (var i = 0; i < list.length; i++) {
			var item = list[i],
				errs = item.Errors || [];
			if (errs.length)
				this.td.log({ failure:"failure to import item: " + JSON.stringify(item) });
		}
		return(result);
	}

	proto._imprules = function (fields) {
		var mappings = [],
			attrmap = {},
			attrs = bundle.load("td-attributes.json") || {},
			result = this.td.ajax({ endpoint:"attributes/custom?componentId=" + this.component + "&appId=" + this.td.appid });
		attrs = attrs[this.name] || {};
		for (var i = 0; i < fields.length; i++)
			mappings.push({ FieldIdentifier:fields[i] });
		if (result.value) {
			// first add "KS:..." attributes to the attrs list
			for (var i = 0; i < result.value.length; i++) {
				var attr = result.value[i];
				if (this._checkattr(attr) && (attr.Name.indexOf("KS:") == 0))
					attrs[attr.Name] = attr.Name.substr(3).toLowerCase().replace(/[^a-z0-9]/g, "");
			}

			// now create the mappings
			for (var i = 0; i < result.value.length; i++) {
				var attr = result.value[i];
				if (this._checkattr(attr) && attrs[attr.Name]) {
					mappings.push({
						FieldIdentifier:attr.Name,
						CustomAttributeID:attr.ID,
						ClearOnNull:false
					});
					attrmap[attr.ID] = attrs[attr.Name];
				}
			}
		}
		return({
				mappings:mappings,
				attrmap:attrmap
			});
	}

	proto._checkattr = function (attr) {
		if (attr && attr.ID && attr.Name) {
			if (!attr.AppID || (attr.AppID == this.td.appid))
				return(true);
		}
		return(false);
	}

	function extend (obj, more, except) {
		if (more) {
			for (var key in more) {
				if (!except || !except[key])
					obj[key] = more[key];
			}
		}
		return(obj);
	}

})(TDTable.prototype);

TDTable.subclass = function (cl, methods) {
	cl.prototype = Object.create(TDTable.prototype);
	cl.prototype.constructor = cl;
	methods(cl.prototype);
}

// #mark -
// #mark TDAssets
// #mark -

function TDAssets (td, keys, screen) {
	TDTable.call(this, td, "assets", 27, keys || [ "ExternalID", "Tag", "SerialNumber", "Name" ], screen);
	this.marker = 1;

	return(this);
}

TDTable.subclass(TDAssets, function (proto) {

	proto.filter = function (base) {
		var filter = base || {};
		filter.SerialLike = filter.SerialLike || "";
		return(filter);
	}

	proto.commit = function (item, update) {
		if (item) {
			item.dirty = this.marker;
			if (update && item.ID) {
				var imp = this._imprules(this.fields),
					prev = this.td.ajax({
						method: "GET",
						endpoint:this.td.appid + "/assets/" + item.ID
					});
				this.td.ajax({
						method: "POST",
						endpoint:this.td.appid + "/assets/" + item.ID,
						data: this._tdify(imp, null, item, prev.value && prev.value.Attributes)
					});
			}
		} else {
			var imp = this._imprules(this.fields),
				items = [];
			app.progress.open(this.list.length + 1);
			for (var i = 0; i < this.list.length; i++) {
				var asset = this.list[i];
				if (asset.ExternalID && asset.dirty && asset.dirty >= this.marker)
					items.push(this._tdify(imp, { ExternalID:asset.ExternalID }, asset));
				asset.dirty = 0;

				if (items.length == 1000) {
					this._importer(imp.mappings, items);
					items = [];
				}
				app.progress.advance();
			}
			if (items.length)
				this._importer(imp.mappings, items);
			app.progress.advance();
			app.progress.close();

			this.marker++;
		}
		return(this);
	}

});

// #mark -
// #mark TDCItems
// #mark -

function TDCItems (td, keys, screen) {
	TDTable.call(this, td, "cmdb", 63, keys || [ "ExternalID", "Name" ], screen);

	return(this);
}

TDTable.subclass(TDCItems, function (proto) {

	proto.filter = function (base) {
		var lookup = this.td.lookup("Configuration Item Types").load(),
			tid = lookup.getid("Software Product", { IsActive:true, AppID:this.td.appid }),
			filter = base || {};

		filter.IsActive = true;
		if (tid)
			filter.TypeIDs = [ tid ];
		return(filter);
	}

	proto.commit = function (item) {
		if (item) {
			var imp = this._imprules(this.fields),
				opts = { endpoint:this.td.appid + "/cmdb" },
				nit = { };
			for (var key in item) {
				if (item.hasOwnProperty(key) && key != "attributes")
					nit[key] = item[key];
			}
			opts.data = this._tdify(imp, nit, item);
			if (item.ID) {
				opts.endpoint += "/" + item.ID;
				opts.method = "PUT";
			}
			var results = this.td.ajax(opts);
			if (!item.ID) {
				nit = results.value;
				if (nit && nit.ID)
					item.ID = nit.ID;
			}
		}
		return(this);
	}

});

// #mark -
// #mark TDLookup
// #mark -

function TDLookup (td, endpoint, key) {
	this.td = td;
	this.endpoint = endpoint;
	this.key = key || "Name";
	this.name2id = {};

	return(this);
}

(function (proto) {

	proto.load = function (force) {
		if (!this.loaded || force) {
			var result = this.td.ajax({
				endpoint: this.endpoint
			});
			if (result.value) {
				this.name2id = {};
				for (var i = 0; i < result.value.length; i++) {
					var item = result.value[i];
					if (item.ID && item[this.key])
						this.name2id[item[this.key]] = item.ID;
				}
				this.loaded = true;
			}
		}

		return(this);
	}

	proto.names = function () {
		var list = [];
		for (var key in this.name2id)
			list.push(key);
		return(list);
	}

	proto.getid = function (name, props) {
		if (!name)
			return(0);

		var item = this.name2id[name];
		if (item)
			return(item);

		if (!this.td || !props)
			return(0);

		var base = {};
		base[this.key] = name;
		var result = this.td.ajax({
				endpoint: this.endpoint,
				data: extend(base, props)
			});

		item = result.value;
		if (!item)
			return(0);

		this.name2id[item[this.key]] = item.ID;
		return(item.ID || 0);
	}

	proto.matchid = function (name) {
		if (!name)
			return(0);

		if (name === "*") {
			for (var key in this.name2id)
				return(this.name2id[key]);
			return(0);
		}

		var item = this.name2id[name];
		if (item)
			return(item);

		var upr = name.toUpperCase();
		for (var key in this.name2id) {
			if (upr.indexOf(key.toUpperCase()) >= 0) {
				this.name2id[name] = this.name2id[key];
				return(this.name2id[key]);
			}
		}

		return(0);
	}

	function extend (obj, more) {
		if (more) {
			for (var key in more)
				obj[key] = more[key];
		}
		return(obj);
	}

})(TDLookup.prototype);
